import json
from functools import wraps

from flask import Blueprint, Response


class Namespace:
    namespaces = []

    def __init__(self, name, url):
        self.blueprint = Blueprint(name, __name__, url_prefix='/api' + url)

    @staticmethod
    def init_namespaces(app):
        for namespace in Namespace.namespaces:
            app.register_blueprint(namespace)

    def route(self, rule, **options):
        def wrapper(func):
            endpoint = options.pop("endpoint", func.__name__)
            self.blueprint.add_url_rule(rule, endpoint, func, **options)
            @wraps(func)
            def inner(*args, **kwargs):
                res, code = func(*args, **kwargs)
                return Response(json.dumps(res), code, mimetype='application/json')
            return inner
        return wrapper

    def register(self):
        Namespace.namespaces.append(self.blueprint)



import os

import pydash
import inflect

from bolinette.cli import cli_env
from bolinette.cli.utils import copy, join, read_manifest, mkdir


def create_model(**kwargs):
    cwd = cli_env['cwd']
    manifest = read_manifest(cwd)
    module = manifest.get('module')
    path = join(cwd, module)
    origin = join(cli_env['origin'], 'files')

    model = kwargs.get('name').lower()
    kwargs['lower'] = model
    kwargs['class'] = pydash.capitalize(model)
    kwargs['plural'] = inflect.engine().plural(model)

    copy(join(origin, 'model.jinja2'), join(path, 'models', f'{model}.py'), kwargs)



from flask import Flask
from flask_cors import CORS
from flask_script import Manager

from bolinette import env, Namespace
from bolinette.database import init_db
from bolinette.jwt import init_jwt
from bolinette.routes import init_routes
from bolinette.scripts import init_commands


class Bolinette:
    def __init__(self, name, **options):
        env_overrides = options.get('env', {})
        self.app = Flask(name, static_url_path='')
        CORS(self.app, supports_credentials=True)
        self.manager = Manager(self.app)
        env.init(self.app, overrides=env_overrides)
        init_jwt(self.app)
        init_db(self.app)
        init_routes(self.app)
        init_commands(self.manager)
        Namespace.init_namespaces(self.app)



from functools import wraps

from flask import request

from bolinette import response, validate
from bolinette.marshalling import get_payload, get_response, marshall
from bolinette.marshalling.definitions import link_foreign_entities


def expects(model, key='default', update=False):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            payload = request.get_json(silent=True) or {}
            def_key = f'{model}.{key}'
            definition = get_payload(def_key)
            if definition is None:
                response.abort(*response.internal_server_error(
                    f'marshalling.unknown_definition:{def_key}'))
            kwargs['payload'] = validate.payload(definition, payload, update)
            link_foreign_entities(definition, payload)
            return func(*args, **kwargs)
        return inner
    return wrapper


def returns(model, key='default', **options):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            as_list = options.get('as_list', False)
            skip_none = options.get('skip_none', False)
            def_key = f'{model}.{key}'
            definition = get_response(def_key)
            if definition is None:
                response.abort(*response.internal_server_error(
                    f'marshalling.unknown_definition:{def_key}'))
            res, code = func(*args, **kwargs)
            if res.get('data') is not None:
                if as_list:
                    res['data'] = [marshall(definition, r, skip_none) for r in res['data']]
                else:
                    res['data'] = marshall(definition, res['data'], skip_none)
            return res, code
        return inner
    return wrapper



from flask_script import Command

from bolinette import db, transaction
import bolinette.models


class InitDb(Command):
    def run(self):
        db.create_all()
        with transaction:
            db.session.add(bolinette.models.Role(name='root'))
            db.session.add(bolinette.models.Role(name='admin'))


class ResetDb(Command):
    def run(self):
        db.drop_all()
        InitDb().run()


def init_commands(manager):
    manager.add_command('initdb', InitDb())
    manager.add_command('resetdb', ResetDb())



from bolinette.marshalling.objects import Field, List, Definition
from bolinette.marshalling.types import types
from bolinette.marshalling.definitions import (
    register, marshall, get_payload, get_response, get_model
)
from bolinette.marshalling.decorators import expects, returns



from bolinette.logger import logger
from bolinette.environment import env
from bolinette.bcrypt import bcrypt
from bolinette.database import db
from bolinette.response import response
from bolinette.namespace import Namespace
from bolinette.transaction import transaction, transactional
from bolinette.mapper import mapper
from bolinette.validator import validate
from bolinette.jwt import jwt
import bolinette.controllers
import bolinette.scripts
from bolinette.bolinette import Bolinette




import os

from flask_sqlalchemy import SQLAlchemy

from bolinette import env, logger

db = SQLAlchemy()


def create_db_uri(_app):
    dbms = env.get('DBMS', 'SQLITE').lower()
    if dbms == 'sqlite':
        return 'sqlite:///' + os.path.join(_app.instance_path,
                                           env.get('SQLITE_FILE', f'{_app.env}.db'))
    if dbms == 'memory':
        return 'sqlite://'
    logger.error(f'Unknown database system "{dbms}"')
    exit(1)


def init_db(app):
    app.config['SQLALCHEMY_DATABASE_URI'] = create_db_uri(app)
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    db.init_app(app)



import os
import random
import string
import subprocess
import re

import yaml
from jinja2 import Template


def mkdir(path):
    os.makedirs(path)


def join(*args):
    return os.path.join(*args)


def write(path, content):
    with open(path, mode='w+') as file:
        file.write(content)


def render(path, params):
    with open(path) as file:
        template = Template(file.read())
        return template.render(**params)


def copy(origin, dest, params):
    content = render(origin, params)
    write(re.sub(r'.jinja2$', '', dest), content)


def random_string(length):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


def read_manifest(path):
    with open(join(path, 'manifest.bolinette.yaml')) as f:
        return yaml.safe_load(f)


def run_shell_command(command, args, out_fn):
    process = subprocess.Popen([command, *args], stdout=subprocess.PIPE)
    for line in iter(process.stdout.readline, ''):
        if len(line) > 0:
            out_fn(line.rstrip().decode('utf-8'))
        else:
            break


def render_directory(origin, dest, params):
    if not os.path.exists(dest):
        mkdir(dest)
    for f in os.listdir(origin):
        if os.path.isdir(join(origin, f)):
            render_directory(join(origin, f), join(dest, f), params)
        if os.path.isfile(join(origin, f)):
            copy(join(origin, f), join(dest, f), params)



from bolinette.cli import cli_env, utils


def create_api(**kwargs):
    cwd = cli_env['cwd']
    origin = utils.join(cli_env['origin'], 'files')
    params = {
        'secret_key': utils.random_string(64),
        'jwt_secret_key': utils.random_string(64),
    }
    utils.render_directory(utils.join(origin, 'api'), cwd, params)
